@@ -17,6 +17,7 @@ import argparse
 from datetime import datetime
 import json
 import os
+import signal
 import ssl
 import sys
 
@@ -468,6 +469,10 @@ def system_resource(identity):
             power_state=app.systems.get_power_state(identity),
             total_memory_gb=try_get(app.systems.get_total_memory),
             bios_version=bios_version,
+            bios_supported=try_get(app.systems.get_bios),
+            processors_supported=try_get(app.systems.get_processors),
+            simple_storage_supported=try_get(
+                app.systems.get_simple_storage_collection),
             total_cpus=try_get(app.systems.get_total_cpus),
             boot_source_target=app.systems.get_boot_device(identity),
             boot_source_mode=try_get(app.systems.get_boot_mode),
@@ -974,6 +979,31 @@ def simple_task():
     return app.render_template('task.json')
 
 
+def cleanup_zombies(signum, frame):
+    """Signal handler for SIGCHLD that reaps all terminated children."""
+    while True:
+        # os.waitpid(-1, os.WNOHANG) checks all children (-1)
+        # without blocking (os.WNOHANG)
+        try:
+            pid, status = os.waitpid(-1, os.WNOHANG)
+        except OSError:
+            # No more children to wait for
+            break
+
+        if pid == 0:
+            # No children have exited yet
+            break
+
+        # Log the reaped zombie process safely to avoid reentrancy issues
+        try:
+            app.logger.debug(
+                "Reaped zombie process %d with status %d", pid, status)
+        except RuntimeError:
+            # If logging fails due to reentrancy, silently continue
+            # The important part is that we reaped the zombie process
+            pass
+
+
 def parse_args():
     parser = argparse.ArgumentParser('sushy-emulator')
     parser.add_argument('--config',
@@ -1035,6 +1065,9 @@ def main():
 
     args = parse_args()
 
+    # Install the SIGCHLD handler to clean up zombie processes
+    signal.signal(signal.SIGCHLD, cleanup_zombies)
+
     app.debug = args.debug
 
     app.configure(config_file=args.config)
