@@ -10,6 +10,8 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
+import os
+import signal
 import tempfile
 from unittest import mock
 
@@ -573,6 +575,191 @@ class BiosTestCase(EmulatorTestCase):
         systems_mock.return_value.reset_bios.assert_called_once_with(self.uuid)
 
 
+class ConditionalResourceAdvertisingTestCase(EmulatorTestCase):
+    """Test that resources are only advertised when driver supports them"""
+
+    def setUp(self):
+        super(ConditionalResourceAdvertisingTestCase, self).setUp()
+        self.set_feature_set("full")
+
+    @patch_resource('indicators')
+    @patch_resource('chassis')
+    @patch_resource('managers')
+    @patch_resource('systems')
+    def test_bios_advertised_when_supported(
+            self, systems_mock, managers_mock, chassis_mock, indicators_mock):
+        """Test BIOS is advertised when driver supports it"""
+        systems_mock = systems_mock.return_value
+        systems_mock.uuid.return_value = 'zzzz-yyyy-xxxx'
+        systems_mock.get_power_state.return_value = 'On'
+        systems_mock.get_boot_device.return_value = 'Cd'
+        systems_mock.get_boot_mode.side_effect = error.NotSupportedError
+        systems_mock.get_bios.return_value = {'attr1': 'value1'}
+        systems_mock.get_processors.side_effect = error.NotSupportedError
+        systems_mock.get_simple_storage_collection.side_effect = (
+            error.NotSupportedError)
+        systems_mock.get_total_memory.side_effect = error.NotSupportedError
+        systems_mock.get_total_cpus.side_effect = error.NotSupportedError
+        systems_mock.get_versions.side_effect = error.NotSupportedError
+        managers_mock.return_value.get_managers_for_system.return_value = []
+        chassis_mock.return_value.chassis = []
+        indicators_mock.return_value.get_indicator_state.return_value = 'Off'
+
+        response = self.app.get('/redfish/v1/Systems/xxxx-yyyy-zzzz')
+
+        self.assertEqual(200, response.status_code)
+        self.assertIn('Bios', response.json)
+        self.assertEqual(
+            '/redfish/v1/Systems/xxxx-yyyy-zzzz/BIOS',
+            response.json['Bios']['@odata.id'])
+
+    @patch_resource('indicators')
+    @patch_resource('chassis')
+    @patch_resource('managers')
+    @patch_resource('systems')
+    def test_bios_not_advertised_when_not_supported(
+            self, systems_mock, managers_mock, chassis_mock, indicators_mock):
+        """Test BIOS is NOT advertised when driver doesn't support it"""
+        systems_mock = systems_mock.return_value
+        systems_mock.uuid.return_value = 'zzzz-yyyy-xxxx'
+        systems_mock.get_power_state.return_value = 'On'
+        systems_mock.get_boot_device.return_value = 'Cd'
+        systems_mock.get_boot_mode.side_effect = error.NotSupportedError
+        systems_mock.get_bios.side_effect = error.NotSupportedError
+        systems_mock.get_processors.side_effect = error.NotSupportedError
+        systems_mock.get_simple_storage_collection.side_effect = (
+            error.NotSupportedError)
+        systems_mock.get_total_memory.side_effect = error.NotSupportedError
+        systems_mock.get_total_cpus.side_effect = error.NotSupportedError
+        systems_mock.get_versions.side_effect = error.NotSupportedError
+        managers_mock.return_value.get_managers_for_system.return_value = []
+        chassis_mock.return_value.chassis = []
+        indicators_mock.return_value.get_indicator_state.return_value = 'Off'
+
+        response = self.app.get('/redfish/v1/Systems/xxxx-yyyy-zzzz')
+
+        self.assertEqual(200, response.status_code)
+        self.assertNotIn('Bios', response.json)
+
+    @patch_resource('indicators')
+    @patch_resource('chassis')
+    @patch_resource('managers')
+    @patch_resource('systems')
+    def test_processors_advertised_when_supported(
+            self, systems_mock, managers_mock, chassis_mock, indicators_mock):
+        """Test Processors is advertised when driver supports it"""
+        systems_mock = systems_mock.return_value
+        systems_mock.uuid.return_value = 'zzzz-yyyy-xxxx'
+        systems_mock.get_power_state.return_value = 'On'
+        systems_mock.get_boot_device.return_value = 'Cd'
+        systems_mock.get_boot_mode.side_effect = error.NotSupportedError
+        systems_mock.get_processors.return_value = [{'id': 'CPU'}]
+        systems_mock.get_bios.side_effect = error.NotSupportedError
+        systems_mock.get_simple_storage_collection.side_effect = (
+            error.NotSupportedError)
+        systems_mock.get_total_memory.side_effect = error.NotSupportedError
+        systems_mock.get_total_cpus.side_effect = error.NotSupportedError
+        systems_mock.get_versions.side_effect = error.NotSupportedError
+        managers_mock.return_value.get_managers_for_system.return_value = []
+        chassis_mock.return_value.chassis = []
+        indicators_mock.return_value.get_indicator_state.return_value = 'Off'
+
+        response = self.app.get('/redfish/v1/Systems/xxxx-yyyy-zzzz')
+
+        self.assertEqual(200, response.status_code)
+        self.assertIn('Processors', response.json)
+        self.assertEqual(
+            '/redfish/v1/Systems/xxxx-yyyy-zzzz/Processors',
+            response.json['Processors']['@odata.id'])
+
+    @patch_resource('indicators')
+    @patch_resource('chassis')
+    @patch_resource('managers')
+    @patch_resource('systems')
+    def test_processors_not_advertised_when_not_supported(
+            self, systems_mock, managers_mock, chassis_mock, indicators_mock):
+        """Test Processors is NOT advertised when driver doesn't support it"""
+        systems_mock = systems_mock.return_value
+        systems_mock.uuid.return_value = 'zzzz-yyyy-xxxx'
+        systems_mock.get_power_state.return_value = 'On'
+        systems_mock.get_boot_device.return_value = 'Cd'
+        systems_mock.get_boot_mode.side_effect = error.NotSupportedError
+        systems_mock.get_processors.side_effect = error.NotSupportedError
+        systems_mock.get_bios.side_effect = error.NotSupportedError
+        systems_mock.get_simple_storage_collection.side_effect = (
+            error.NotSupportedError)
+        systems_mock.get_total_memory.side_effect = error.NotSupportedError
+        systems_mock.get_total_cpus.side_effect = error.NotSupportedError
+        systems_mock.get_versions.side_effect = error.NotSupportedError
+        managers_mock.return_value.get_managers_for_system.return_value = []
+        chassis_mock.return_value.chassis = []
+        indicators_mock.return_value.get_indicator_state.return_value = 'Off'
+
+        response = self.app.get('/redfish/v1/Systems/xxxx-yyyy-zzzz')
+
+        self.assertEqual(200, response.status_code)
+        self.assertNotIn('Processors', response.json)
+
+    @patch_resource('indicators')
+    @patch_resource('chassis')
+    @patch_resource('managers')
+    @patch_resource('systems')
+    def test_simple_storage_advertised_when_supported(
+            self, systems_mock, managers_mock, chassis_mock, indicators_mock):
+        """Test SimpleStorage is advertised when driver supports it"""
+        systems_mock = systems_mock.return_value
+        systems_mock.uuid.return_value = 'zzzz-yyyy-xxxx'
+        systems_mock.get_power_state.return_value = 'On'
+        systems_mock.get_boot_device.return_value = 'Cd'
+        systems_mock.get_boot_mode.side_effect = error.NotSupportedError
+        systems_mock.get_simple_storage_collection.return_value = (
+            {'virtio': {}})
+        systems_mock.get_bios.side_effect = error.NotSupportedError
+        systems_mock.get_processors.side_effect = error.NotSupportedError
+        systems_mock.get_total_memory.side_effect = error.NotSupportedError
+        systems_mock.get_total_cpus.side_effect = error.NotSupportedError
+        systems_mock.get_versions.side_effect = error.NotSupportedError
+        managers_mock.return_value.get_managers_for_system.return_value = []
+        chassis_mock.return_value.chassis = []
+        indicators_mock.return_value.get_indicator_state.return_value = 'Off'
+
+        response = self.app.get('/redfish/v1/Systems/xxxx-yyyy-zzzz')
+
+        self.assertEqual(200, response.status_code)
+        self.assertIn('SimpleStorage', response.json)
+        self.assertEqual(
+            '/redfish/v1/Systems/xxxx-yyyy-zzzz/SimpleStorage',
+            response.json['SimpleStorage']['@odata.id'])
+
+    @patch_resource('indicators')
+    @patch_resource('chassis')
+    @patch_resource('managers')
+    @patch_resource('systems')
+    def test_simple_storage_not_advertised_when_not_supported(
+            self, systems_mock, managers_mock, chassis_mock, indicators_mock):
+        """Test SimpleStorage NOT advertised when driver doesn't support"""
+        systems_mock = systems_mock.return_value
+        systems_mock.uuid.return_value = 'zzzz-yyyy-xxxx'
+        systems_mock.get_power_state.return_value = 'On'
+        systems_mock.get_boot_device.return_value = 'Cd'
+        systems_mock.get_boot_mode.side_effect = error.NotSupportedError
+        systems_mock.get_simple_storage_collection.side_effect = (
+            error.NotSupportedError)
+        systems_mock.get_bios.side_effect = error.NotSupportedError
+        systems_mock.get_processors.side_effect = error.NotSupportedError
+        systems_mock.get_total_memory.side_effect = error.NotSupportedError
+        systems_mock.get_total_cpus.side_effect = error.NotSupportedError
+        systems_mock.get_versions.side_effect = error.NotSupportedError
+        managers_mock.return_value.get_managers_for_system.return_value = []
+        chassis_mock.return_value.chassis = []
+        indicators_mock.return_value.get_indicator_state.return_value = 'Off'
+
+        response = self.app.get('/redfish/v1/Systems/xxxx-yyyy-zzzz')
+
+        self.assertEqual(200, response.status_code)
+        self.assertNotIn('SimpleStorage', response.json)
+
+
 @patch_resource('systems')
 class EthernetInterfacesTestCase(EmulatorTestCase):
 
@@ -1037,3 +1224,163 @@ class TaskServiceTestCase(EmulatorTestCase):
                          response.json['@odata.id'])
         self.assertEqual('Task 42', response.json['Name'])
         self.assertEqual('Completed', response.json['TaskState'])
+
+
+class ZombieCleanupTestCase(base.BaseTestCase):
+    """Test case for zombie process cleanup functionality"""
+
+    def setUp(self):
+        super().setUp()
+        # Store original signal handler to restore later
+        self.original_sigchld_handler = signal.signal(
+            signal.SIGCHLD, signal.SIG_DFL)
+
+    def tearDown(self):
+        # Restore original signal handler
+        signal.signal(signal.SIGCHLD, self.original_sigchld_handler)
+        super().tearDown()
+
+    @mock.patch('os.waitpid')
+    def test_cleanup_zombies_single_child(self, mock_waitpid):
+        """Test that cleanup_zombies reaps a single zombie child"""
+        # Mock waitpid to return a single child that has exited
+        # First call returns child, second returns no more children
+        mock_waitpid.side_effect = [(12345, 0), (0, 0)]
+
+        with mock.patch.object(main.app, 'logger') as mock_logger:
+            main.cleanup_zombies(signal.SIGCHLD, None)
+
+        # Verify waitpid was called with correct parameters
+        expected_calls = [
+            mock.call(-1, os.WNOHANG),
+            mock.call(-1, os.WNOHANG)
+        ]
+        mock_waitpid.assert_has_calls(expected_calls)
+
+        # Verify the logging statement was called with correct message
+        # Note: In real usage, logging might fail due to reentrancy
+        mock_logger.debug.assert_called_once_with(
+            "Reaped zombie process %d with status %d", 12345, 0)
+
+    @mock.patch('os.waitpid')
+    def test_cleanup_zombies_multiple_children(self, mock_waitpid):
+        """Test that cleanup_zombies reaps multiple zombie children"""
+        # Mock waitpid to return multiple children
+        mock_waitpid.side_effect = [
+            (12345, 0),    # First child
+            (12346, 256),  # Second child (exit code 1)
+            (0, 0)         # No more children
+        ]
+
+        with mock.patch.object(main.app, 'logger') as mock_logger:
+            main.cleanup_zombies(signal.SIGCHLD, None)
+
+        # Verify waitpid was called three times
+        self.assertEqual(mock_waitpid.call_count, 3)
+
+        # Verify both logging statements were called
+        expected_calls = [
+            mock.call("Reaped zombie process %d with status %d", 12345, 0),
+            mock.call("Reaped zombie process %d with status %d", 12346, 256)
+        ]
+        mock_logger.debug.assert_has_calls(expected_calls)
+
+    @mock.patch('os.waitpid')
+    def test_cleanup_zombies_no_children(self, mock_waitpid):
+        """Test that cleanup_zombies handles no zombie children gracefully"""
+        # Mock waitpid to return no children
+        mock_waitpid.return_value = (0, 0)
+
+        with mock.patch.object(main.app, 'logger') as mock_logger:
+            main.cleanup_zombies(signal.SIGCHLD, None)
+
+        # Verify waitpid was called once
+        mock_waitpid.assert_called_once_with(-1, os.WNOHANG)
+
+        # Verify no logging statements were made
+        mock_logger.debug.assert_not_called()
+
+    @mock.patch('os.waitpid')
+    def test_cleanup_zombies_oserror_handling(self, mock_waitpid):
+        """Test that cleanup_zombies handles OSError gracefully"""
+        # Mock waitpid to raise OSError (no more children)
+        mock_waitpid.side_effect = OSError("No child processes")
+
+        with mock.patch.object(main.app, 'logger') as mock_logger:
+            main.cleanup_zombies(signal.SIGCHLD, None)
+
+        # Verify waitpid was called once
+        mock_waitpid.assert_called_once_with(-1, os.WNOHANG)
+
+        # Verify no logging statements were made
+        mock_logger.debug.assert_not_called()
+
+    @mock.patch('signal.signal')
+    def test_signal_handler_installation(self, mock_signal):
+        """Test that the SIGCHLD signal handler is properly installed"""
+        # Create a mock args object with all necessary attributes
+        mock_args = mock.Mock()
+        mock_args.ssl_certificate = None
+        mock_args.ssl_key = None
+
+        # Mock the signal.signal call to verify it's called correctly
+        with mock.patch('sushy_tools.emulator.main.parse_args') as mock_parse:
+            mock_parse.return_value = mock_args
+
+            # Mock app configuration to avoid actual Flask app setup
+            with mock.patch('sushy_tools.emulator.main.app') as mock_app:
+                mock_app.configure = mock.Mock()
+                mock_app.config = {}
+                mock_app.run = mock.Mock()
+
+                # Call main function
+                main.main()
+
+        # Verify signal.signal was called with SIGCHLD and cleanup function
+        mock_signal.assert_called_with(signal.SIGCHLD, main.cleanup_zombies)
+
+    @mock.patch('os.waitpid')
+    def test_cleanup_zombies_with_different_exit_statuses(self, mock_waitpid):
+        """Test that cleanup_zombies handles different exit statuses"""
+        # Mock waitpid to return children with different exit statuses
+        mock_waitpid.side_effect = [
+            (12345, 0),      # Normal exit
+            (12346, 256),    # Exit code 1 (256 = 1 << 8)
+            (12347, 512),    # Exit code 2 (512 = 2 << 8)
+            (0, 0)           # No more children
+        ]
+
+        with mock.patch.object(main.app, 'logger') as mock_logger:
+            main.cleanup_zombies(signal.SIGCHLD, None)
+
+        # Verify all three children were reaped
+        expected_calls = [
+            mock.call("Reaped zombie process %d with status %d", 12345, 0),
+            mock.call("Reaped zombie process %d with status %d", 12346, 256),
+            mock.call("Reaped zombie process %d with status %d", 12347, 512)
+        ]
+        mock_logger.debug.assert_has_calls(expected_calls)
+
+    @mock.patch('os.waitpid')
+    def test_cleanup_zombies_logging_failure_handled(self, mock_waitpid):
+        """Test that cleanup_zombies handles logging failures gracefully"""
+        # Mock waitpid to return a child
+        mock_waitpid.side_effect = [(12345, 0), (0, 0)]
+
+        with mock.patch.object(main.app, 'logger') as mock_logger:
+            # Make logging raise an exception (simulating reentrancy issue)
+            mock_logger.debug.side_effect = RuntimeError("Logging reentrancy")
+
+            # This should not raise an exception
+            main.cleanup_zombies(signal.SIGCHLD, None)
+
+        # Verify waitpid was still called correctly
+        expected_calls = [
+            mock.call(-1, os.WNOHANG),
+            mock.call(-1, os.WNOHANG)
+        ]
+        mock_waitpid.assert_has_calls(expected_calls)
+
+        # Verify logging was attempted
+        mock_logger.debug.assert_called_once_with(
+            "Reaped zombie process %d with status %d", 12345, 0)
